import json
import os
from collections.abc import Callable
from pathlib import Path
import bayesbuilding.models as mods
import warnings

import arviz as az
import numpy as np
import pandas as pd
import pymc as pm


def r2_score(y_true, y_pred):
    numerator = ((y_true - y_pred) ** 2).sum(axis=0)
    denominator = ((y_true - np.average(y_true, axis=0)) ** 2).sum(axis=0)
    return 1 - numerator / denominator


def custom_serializer(obj):
    if callable(obj) and hasattr(obj, "__name__"):
        return obj.__name__
    else:
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


class PymcWrapper:
    """
    A class simplifying the creation of probabilistic models using PyMC3.

    This class abstracts some complexities of PyMC while enforcing a structured
     framework:
    - Features are provided as pandas DataFrames.
    - Only a single target variable, provided as a pandas Series, is allowed.
    - The model is defined by a model function, which returns deterministic outputs.
    - The likelihood function is a Normal distribution, with the mean given by the
        output of the model function and the standard deviation (sigma) being
        a variable.
    - For change point models, sigma may have multiple dimensions.
    - The likelihood function is used for prediction, with the variable
        named "observations".

    Parameters:
    -----------
    model_function : Callable
        A function defining the PyMC model. It takes two arguments: 'x' and
        'variable_dict'. 'x' is a 2D array derived from the features DataFrame.
        Use standard indexing as DataFrame column names cannot be used within
        the function's scope.
        'variable_dict' is a dictionary where keys are variable names and values
        are PyMC variables.
    priors_dict : dict[str:(Callable, dict)]
        A dictionary containing prior distributions for model parameters.
        Keys are parameter names, and values are tuples where the first element is
        a PyMC callable defining the prior distribution, and the second element is
        a dictionary of parameters for the prior distribution.
        The dictionary must include a variable called 'sigma', representing the
        likelihood standard deviation.
    sigma_change_point_idx : int, optional
        Index of 'x' in the 2nd dimension (feature axis) indicating the change point
        for sigma in change point models.

    Attributes:
    -----------
    var_names : list
        Names of the model variables.
    features_names : list
        Names of the features used in the model. Only available after performing
        prior prediction or sampling.
    target_name : str
        Name of the target variable. Only available after sampling.
    trace : InferenceData
        PyMC trace containing samples.
    sigma_change_point_idx : int, optional
        Index indicating the change point for sigma.
    _observations : pm.Normal
        PyMC3 Normal distribution defining the model's likelihood.
    _data_dict : dict
        Dictionary containing PyMC3 MutableData objects.
    _variables_dict : dict
        Dictionary containing PyMC3 variables for the model parameters.

    Methods:
    --------
    __repr__():
        Returns a string representation of the model's variables, features, and target.
    sample_prior():
        Samples from the prior predictive distribution.
    sample():
        Samples to obtain posterior distributions using the NUTS algorithm.
    sample_posterior_predictive():
        Samples from the posterior predictive distribution.
    get_summary():
        Returns a summary of the posterior distribution. Use the 'group' argument to
        select among 'prior', 'prior_predictive', 'posterior', or
        'posterior_predictive'.
    get_loo_score():
        Computes the leave-one-out cross-validation (LOO) score.
    save_model(Path):
        Save the model to the required path. It will contain 3 traces
        and a json file.
    load_model(Path):
        Load the model traces and json file contained in the desired path and build
        the pymc model.
        If the loaded model function is not in bayesbuilding.model, it must
        be provided separately and the pymc model must be build using the build_model()
        method.
    plot_dist_comparison(var_names):
        Plot compare prior and posterior distributions of variables and observations
        var_names arguments filter the values to display. default is self.var_names
    """

    def __init__(
        self,
        model_function: Callable = None,
        priors_dict: dict[str:(Callable, dict)] = None,
        sigma_change_point_idx=None,
    ):
        self.model_function = model_function
        self.priors_dict = priors_dict
        self.features_names = None
        self.target_name = None
        self.var_names = None
        self.model = None
        self.sigma_change_point_idx = sigma_change_point_idx
        self.traces = {
            "prior": az.data.inference_data.InferenceData(),
            "sampling": az.data.inference_data.InferenceData(),
            "posterior": az.data.inference_data.InferenceData(),
        }
        self._x_train = None
        self._y_train = None
        self._observations = None
        self._data_dict = None
        self._variables_dict = None

        self.build_model()

    def __repr__(self):
        string_out = """"""
        string_out += """=== Variables names : \n"""
        if self.var_names is not None:
            for var in self.var_names:
                string_out += f"""- {var} \n"""
        string_out += """\n"""

        string_out += """=== Features names: \n"""
        if self.features_names is not None:
            for feat in self.features_names:
                string_out += f"""- {feat} \n"""
        string_out += """\n"""

        string_out += """=== Target name: \n"""
        if self.target_name is not None:
            string_out += f"""- {self.target_name} \n"""

        return string_out

    def save_model(self, dir_path: Path):
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        for name, traces in self.traces.items():
            traces.to_netcdf((dir_path / f"{name}.nc").as_posix())

        with open(dir_path / "config.json", "w", encoding="utf-8") as f:
            to_dump = {
                "model_function": self.model_function,
                "priors_dict": self.priors_dict,
                "sigma_change_point_idx": self.sigma_change_point_idx,
                "features_names": self.features_names,
                "target_name": self.target_name,
            }
            json.dump(
                to_dump, f, ensure_ascii=False, default=custom_serializer, indent=4
            )

    def load_model(self, dir_path: Path):
        if not os.path.exists(dir_path):
            raise ValueError(f"Provided dir_path : {dir_path} not found")

        for name, _ in self.traces.items():
            self.traces[name] = az.from_netcdf((dir_path / f"{name}.nc").as_posix())

        with open(dir_path / "config.json", encoding="utf-8") as f:
            config_dict = json.load(f)

        for attr, value in config_dict.items():
            setattr(self, attr, value)

        for val in self.priors_dict.values():
            val[0] = getattr(pm, val[0])

        if self.model_function is not None:
            try:
                self.model_function = getattr(mods, self.model_function)
            except AttributeError:
                warnings.warn(
                    f"Model function {self.model_function} not found in"
                    f"bayesbuilding.models. Load a model function before running"
                    f"build_model() method"
                )
                self.model_function = None

        self.build_model()

    def build_model(self):
        if self.model_function is not None and self.priors_dict is not None:
            self.model = pm.Model()
            with self.model:
                # Set empty data objects
                self._data_dict = {
                    "x": pm.MutableData(
                        name="x", value=np.array([[]]), dims=["date", "features"]
                    ),
                    "y": pm.MutableData(name="y", value=np.array([]), dims=["target"]),
                }

                # Set priors
                self._variables_dict = {
                    name: val[0](**val[1]) for name, val in self.priors_dict.items()
                }

                mu = pm.Deterministic(
                    name="mu",
                    var=self.model_function(self._data_dict["x"], self._variables_dict),
                )

                # Combine into likelihood function
                sigma = (
                    self._variables_dict["sigma"][
                        self._data_dict["x"][:, self.sigma_change_point_idx].astype(int)
                    ]
                    if self.sigma_change_point_idx is not None
                    else self._variables_dict["sigma"]
                )

                self._observations = pm.Normal(
                    name="observations",
                    mu=mu,
                    sigma=sigma,
                    observed=self._data_dict["y"],
                    shape=self._data_dict["x"].shape[0],
                )
            self.var_names = list(self.priors_dict.keys())

    def sample_prior(
        self,
        samples: int = 500,
        x: pd.DataFrame = None,
        var_names=None,
        sample_kwargs=None,
    ):
        if sample_kwargs is None:
            sample_kwargs = {}
        with self.model:
            if x is not None:
                pm.set_data({"x": x})
                self.features_names = list(x.columns)

            self.traces["prior"] = pm.sample_prior_predictive(
                samples=samples, var_names=var_names, **sample_kwargs
            )

    def sample(
        self,
        y: pd.Series,
        x: pd.DataFrame = None,
        draws: int = 1000,
        tune: int = 1000,
        chains=None,
        sample_kwargs=None,
    ):
        if sample_kwargs is None:
            sample_kwargs = {}

        with self.model:
            if x is not None:
                pm.set_data({"x": x})
                self.features_names = list(x.columns)

            pm.set_data({"y": y})
            self.target_name = y.name

            self.traces["sampling"] = pm.sample(
                draws=draws, tune=tune, chains=chains, **sample_kwargs
            )

        self.sample_posterior_predictive()

        self._x_train = x
        self._y_train = y

    def sample_posterior_predictive(self, x: pd.DataFrame = None, sample_kwargs=None):
        if self.traces["sampling"] is None:
            raise ValueError(
                "No posterior trace available. Perform sampling before"
                "sampling from posterior"
            )

        if sample_kwargs is None:
            sample_kwargs = {}

        with self.model:
            if x is not None:
                pm.set_data({"x": x})

            self.traces["posterior"] = pm.sample_posterior_predictive(
                trace=self.traces["sampling"],
                var_names=["observations"],
                **sample_kwargs,
            )

    def get_summary(
        self, group: str = None, var_names=None, filter_vars=None, summary_kwargs=None
    ):
        if group not in self.traces.keys():
            raise ValueError(
                f"Unknown group {group} choose one of {self.traces.keys()}"
            )
        if summary_kwargs is None:
            summary_kwargs = {}

        return az.summary(
            data=self.traces[group],
            var_names=var_names,
            filter_vars=filter_vars,
            **summary_kwargs,
        )

    def get_loo_score(self, loo_kwargs=None):
        if self.traces["sampling"] is None:
            raise ValueError(
                "Sampling trace is not available, perform sampling "
                "before checking loo score"
            )
        if loo_kwargs is None:
            loo_kwargs = {}
        with self.model:
            pm.set_data({"x": self._x_train, "y": self._y_train})
            pm.compute_log_likelihood(self.traces["sampling"])
        return az.loo(data=(self.traces["sampling"]), **loo_kwargs)

    def score(
        self,
        x: pd.DataFrame,
        y: pd.Series,
        score_function: Callable = r2_score,
        sample_kwargs: dict = None,
    ):
        if sample_kwargs is None:
            sample_kwargs = {}

        self.sample_posterior_predictive(x=x, sample_kwargs=sample_kwargs)
        post_trace = self.traces["posterior"].posterior_predictive["observations"]
        flattened_trace = np.array(post_trace).reshape(-1, post_trace.shape[-1])

        scores_array = np.array(
            [score_function(y, sample) for sample in flattened_trace]
        )
        return {"mean_score": np.mean(scores_array), "sd_score": np.std(scores_array)}

    def plot_dist_comparison(self, var_names: list[str] = None, plot_dist_kwargs=None):
        if var_names is None:
            var_names = self.var_names
        if plot_dist_kwargs is None:
            plot_dist_kwargs = {}
        temp = az.data.inference_data.InferenceData()
        temp.extend(self.traces["sampling"])
        temp.extend(self.traces["prior"])
        return az.plot_dist_comparison(temp, var_names=var_names, **plot_dist_kwargs)



import pymc as pm


def occ_cp(x, variables_dict: dict):
    """
    Model a system assuming it has two distinct fairly constant behaviour
    during specified period. For example week-days and weekends, or holidays.
    :param x: The features. x[:, 0] must be a columns of boolean values, or integer
    indicating the period (from 0, to n period)
    :param variables_dict: variable dictionary with a unique distribution called
    set_point of shape n period
    :return:
    """
    wd_we = x[:, 0].astype(int)
    set_point = variables_dict["set_point"]
    return set_point[wd_we]


def season_cp_occ_cp_heating_cooling_es(x, variables_dict: dict):
    """
    Season Occupation Change point Heating / Cooling Energy Signature
    source S. Rouchier (https://buildingenergygeeks.org/bayesianmv.html)
    Piecewise linear model to predict the overall building energy consumption.
    Assume 3 distinct periods Heating, Cooling, mid-season:
    n distinct functions for n occupation typology.

    if tay_heat > t_ext :
        E = g_{heat} * (tau_{heat} - t_{ext}) + base

    if tau_cool > t_ext :
        E = g_{cool} * (t_{ext} - tau_{cool}) + base

    else :
        E = base

    :param x: The features . x[:, 0] must be a columns of boolean values, or integer
    indicating the period (from 0, to n period), x[:, 1] must be external temperatures
    :param variables_dict: mandatory model variables are : "base", "g_h", "g_c",
    "tau_h", "tau_c"
    :return: Energy consumption
    """
    occupation = x[:, 0].astype(int)
    t_ext = x[:, 1]

    base = variables_dict["base"]
    g_h = variables_dict["g_h"]
    g_c = variables_dict["g_c"]
    tau_h = variables_dict["tau_h"]
    tau_c = variables_dict["tau_c"]

    baseline = base[occupation]
    heat = g_h[occupation] * pm.math.maximum(tau_h[occupation] - t_ext, 0)
    cool = g_c[occupation] * pm.math.maximum(t_ext - tau_c[occupation], 0)
    return baseline + heat + cool


def season_cp_heating_es(x, variable_dict):
    """
    Seasonal Change Point Heating Energy Signature

    During winter the overall building energy consumption is modeled as a linear
    function : Text * G + baseline. Where G is the overall heat loss [kWh/°C]
    and baseline is the "process" energy consumption.
    During summer, only baseline remains.
    The changepoint tau is based on the exterior temperature

    :param x: single column 2D array. x[:, 0] is the external air temperature
    :param variable_dict: mandatory model variables are : "g", "tau", "base"
    :return: overall building consumption
    """
    t_ext = x[:, 0]
    g = variable_dict["g"]
    tau = variable_dict["tau"]
    baseline = variable_dict["base"]

    consumption = g * pm.math.maximum(tau - t_ext, 0)
    return consumption + baseline


def season_cp_occ_cp_heating_es(x, variable_dict):
    """
    Seasonal Change Point Heating Energy Signature

    During winter the overall building energy consumption is modeled as a linear
    function : Text * G + baseline. Where G is the overall heat loss [kWh/°C]
    and baseline is the "process" energy consumption.
    During summer, only baseline remains.
    The changepoint tau is based on the exterior temperature
    n distinct function for n occupation typologie

    :param x: 2D array. x[:, 0] is the occupation index, x[:, 1] is the esternal
    air temperature
    :param variable_dict: mandatory model variables are : "g", "tau", "base". Each
    variables have a shape of diemension n.
    :return: energy consumption
    """
    occ = x[:, 0].astype(int)
    t_ext = x[:, 1]
    g = variable_dict["g"]
    tau = variable_dict["tau"]
    baseline = variable_dict["baseline"]

    consumption = g[occ] * pm.math.maximum(tau[occ] - t_ext, 0)
    return consumption + baseline[occ]


def we_cst_wd_radiation_lighting(x, variable_dict):
    """
    Artificial Lighting energy consumption model.
    Depending on the occupation, returns a baseline_we consumption (weekends, holidays),
    or an external radiation dependant term + baseline_wd

    :param variable_dict: mandatory model variables are :
    - base_we: baseline with no occupation
    - base_wd: baseline during weekdays
    - fs: coefficient to account for natural lighting based on solar radiations
    :param x: x[:, 0] is boolean series or 0-1 describing 2 occupation behaviour,
        x[:, 1] is solar radiation. For solar radiation, use Global horizontal,
        or custom projection.
    """

    fs = variable_dict["fs"]
    base_we = variable_dict["base_we"]
    base_wd = variable_dict["base_wd"]

    return pm.math.switch(x[:, 0], base_we, base_wd + fs * x[:, 1])


def season_cp_occ_cp_rad_heating_cooling_es(x, variables_dict: dict):
    """
    Add radiation to Season Occupation Change point Heating / Cooling Energy Signature
    source S. Rouchier (https://buildingenergygeeks.org/bayesianmv.html)
    Piecewise linear model to predict the overall building energy consumption.
    Assume 3 distinct periods Heating, Cooling, mid-season:
    n distinct functions for n occupation typology.

    if tau_heat > t_ext :
        heat = g_{heat} * (tau_{heat} - t_{ext})

    if tau_cool < t_ext :
        cool = g_{cool} * (t_{ext} - tau_{cool})

    if tau_rad_h > rad:
        solar_heat = fs_{heat} * (taurad_{heat} - rad)

    if tau_rad_c < rad:
        solar_cool = - fs_{heat} * (taurad_{cool} - rad)

    E = base

    :param x: The features . x[:, 0] must be a columns of boolean values, or integer
    indicating the period (from 0, to n period), x[:, 1] must be external temperatures,
    x[:, 2] is a measure of solar radiation. For exemple projected radiation or
    Global Horizontal depending on the value and the meaning of fs
    :param variables_dict: mandatory model variables are : "base", "g_h", "g_c",
    "tau_h", "tau_c", "fs_h", "fs_c", "tau_rad_h", "tau_rad_c".
    :return: Energy consumption
    """
    occupation = x[:, 0].astype(int)
    t_ext = x[:, 1]
    rad = x[:, 2]

    base = variables_dict["base"]
    g_h = variables_dict["g_h"]
    g_c = variables_dict["g_c"]
    tau_h = variables_dict["tau_h"]
    tau_c = variables_dict["tau_c"]
    fs_h = variables_dict["fs_h"]
    fs_c = variables_dict["fs_c"]
    tau_rad_h = variables_dict["tau_rad_h"]
    tau_rad_c = variables_dict["tau_rad_c"]

    baseline = base[occupation]
    solar_heat = fs_h[occupation] * pm.math.maximum(tau_rad_h[occupation] - rad, 0)
    solar_cool = -fs_c[occupation] * pm.math.maximum(rad - tau_rad_c[occupation], 0)
    heat = g_h[occupation] * pm.math.maximum(tau_h[occupation] - t_ext, 0)
    cool = g_c[occupation] * pm.math.maximum(t_ext - tau_c[occupation], 0)
    return baseline + heat + cool + solar_cool + solar_heat


def ppv_projected_rad_cst_eff(x, variables_dict: dict):
    """
    Simplest model of pv panels. Constant efficiency.
    Radiation are provided as kWh/m² and must already be projected in the plan
    of the pannel
    :param x: The features . x[:, 0] is the solar radiation in Wh/m² or kWh/m²
    :param variables_dict: mandatory model variables are : "surface" and "efficiency"
    :return:  surface * efficiency * radiations
    """
    rad = x[:, 0]
    efficiency = variables_dict["efficiency"]
    surface = variables_dict["surface"]

    return surface * efficiency * rad



from pathlib import Path

import numpy as np
import pandas as pd
import pymc as pm

from bayesbuilding.models import season_cp_heating_es
from bayesbuilding.plotting import time_series_hdi, changepoint_graph
from bayesbuilding.wrapper import PymcWrapper


class TestWrapper:
    def test_pymc_wrapper(self, tmp_path):
        data = pd.DataFrame(
            {
                "Text": [
                    7.149576,
                    6.622098,
                    11.563192,
                    13.513255,
                    17.170108,
                    22.034521,
                    21.532960,
                    22.664882,
                    21.892324,
                    17.477688,
                    11.189259,
                    8.383155,
                    7.207740,
                    10.597761,
                ]
            },
            index=pd.date_range("2023-01", freq="ME", periods=14),
        )

        true_g = 50  # kWh/°C
        true_base = 50  # kWh
        true_tau = 14  # °C
        true_sigma = 10

        np.random.seed(42)
        noise = np.random.randn(14)
        noise_g = true_sigma * noise + true_g
        noise_base = true_sigma * noise + true_base

        data["heating"] = noise_g * np.maximum(true_tau - data["Text"], 0) + noise_base

        data_train = data.loc["2023", :]
        data_test = data.loc["2024", :]

        test_model = PymcWrapper(
            model_function=season_cp_heating_es,
            priors_dict={
                "g": (pm.Normal, dict(name="g", mu=40, sigma=5)),
                "tau": (pm.Normal, dict(name="tau", mu=12, sigma=1)),
                "base": (pm.Normal, dict(name="base", mu=30, sigma=5)),
                "sigma": (pm.Normal, dict(name="sigma", mu=12, sigma=1)),
            },
        )

        test_model.sample_prior(
            samples=4000, x=data_train[["Text"]], sample_kwargs={"random_seed": 42}
        )

        assert round(test_model.get_summary(group="prior").loc["g", "mean"], 1) == 40.1
        assert round(test_model.get_summary(group="prior").loc["g", "sd"], 1) == 5.0

        test_model.sample(
            x=data_train[["Text"]],
            y=data_train["heating"],
            draws=4000,
            sample_kwargs={"random_seed": 42},
        )

        assert (
            round(test_model.get_summary(group="sampling").loc["g", "mean"], 1) == 47.6
        )
        assert round(test_model.get_summary(group="sampling").loc["g", "sd"], 1) == 1.9

        score_res = test_model.score(
            x=data_test[["Text"]],
            y=data_test["heating"],
            sample_kwargs={"random_seed": 42},
        )

        score_res = {key: round(val, 2) for key, val in score_res.items()}
        assert score_res == {"mean_score": 0.76, "sd_score": 0.07}

        test_model.get_loo_score()

        # === test save / load ===
        test_model.save_model(Path(tmp_path))

        new_model = PymcWrapper()
        new_model.load_model(Path(tmp_path))

        # === test plots ===
        test_model.sample_posterior_predictive(data[["Text"]])

        test_model.plot_dist_comparison()

        time_series_hdi(
            measure_ts=data["heating"],
            prediction=test_model.traces["posterior"].posterior_predictive[
                "observations"
            ],
            title="Posterior model accuracy",
            y_label="Energy consumption [kWh]",
            backend="plotly",
            image_path=Path(
                r"C:\Users\bdurandestebe\Documents\49_BIM_GEM_E\cta_env\test.png"
            ),
        )
        time_series_hdi(
            measure_ts=data["heating"],
            prediction=test_model.traces["posterior"].posterior_predictive[
                "observations"
            ],
            title="Posterior model accuracy",
            y_label="Energy consumption [kWh]",
            backend="matplotlib",
        )

        changepoint_graph(
            data["Text"],
            data["heating"],
            test_model.traces["posterior"].posterior_predictive["observations"],
            backend="plotly",
            x_label="text",
            y_label="heating",
            title="test",
        )

        changepoint_graph(
            data["Text"],
            data["heating"],
            test_model.traces["posterior"].posterior_predictive["observations"],
            backend="matplotlib",
            x_label="text",
            y_label="heating",
            title="test",
        )



import numpy as np
import pandas as pd

from bayesbuilding.wrapper import PymcWrapper


class LooSelector:
    def __init__(self, model_dict: dict[str, PymcWrapper]):
        self.model_dict = model_dict

    def sample(
        self,
        y: pd.Series,
        x: pd.DataFrame = None,
        draws: int = 1000,
        tune: int = 1000,
        chains=None,
        sample_kwargs=None,
        models: list["str"] = None,
    ):
        if models is None:
            models = self.model_dict.keys()
        elif np.any(
            ~np.array(
                [
                    True if mod in list(self.model_dict.keys()) else False
                    for mod in models
                ]
            )
        ):
            raise ValueError(
                "model name in models argument not found in model_dict keys"
            )



"""BayesBuilding"""

from setuptools import setup, find_packages

# Get the long description from the README file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="bayesbuilding",
    version="0.1.0",
    description="Bayesian approach for building energy modeling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BuildingEnergySimulationTools/bayesbuilding",
    author="Nobatek/INEF4",
    author_email="bdurandestebe@nobatek.inef4.com",
    license="License :: OSI Approved :: BSD License",
    # keywords=[
    # ],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.10",
    install_requires=[
        "pymc>=5.10.3",
        "pandas>=1.3.4",
        "numpy>=1.26.3",
        "arviz>=0.18.0",
        "matplotlib>=3.7.5",
        "plotly>=5.13.1",
    ],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
)



import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import xarray
from plotly.subplots import make_subplots
from pathlib import Path


def get_quantiles(prediction, lower_q, upper_q, lower_cut, upper_cut):
    if isinstance(prediction, xarray.core.dataarray.DataArray):
        prediction = np.array(prediction)

    if prediction.ndim > 2:  # Assume it's because we have several chains
        prediction = prediction.reshape((-1, prediction.shape[2]))

    prediction_q = np.quantile(
        prediction,
        q=[lower_q, 0.5, upper_q],
        axis=0,
    )
    if lower_cut is not None:
        prediction_q[prediction_q < lower_cut] = lower_cut

    if upper_cut is not None:
        prediction_q[prediction_q < upper_cut] = upper_cut

    return prediction_q


def time_series_hdi(
    measure_ts: pd.Series,
    prediction: np.ndarray | xarray.core.dataarray.DataArray,
    y_label: str = None,
    title: str = None,
    lower_q=0.025,
    upper_q=0.975,
    upper_cut=None,
    lower_cut=None,
    image_path: Path = None,
    figsize: tuple = (10, 6),
    backend: str = "plotly",
):
    """
    Visualise actual measure time series  and probabilist model prediction.
    Measure are plotted as scatter points, predictions are plotted using a surface
    bounded by lower and upper quantiles around the median.

    Parameters:
    - measure_ts (pd.Series): The observed time series data.
    - prediction (np.ndarray): Array containing predictions, potentially from multiple
    chains.
    - y_label (str): Label for the y-axis of the plot.
    - title (str): Title for the plot.
    - lower_q (float): Lower quantile for the high-density interval (default is 0.025).
    - upper_q (float): Upper quantile for the high-density interval (default is 0.975).
    - upper_cut (float): Upper bound for the high-density interval (optional).
    - lower_cut (float): Lower bound for the high-density interval (optional).
    - backend (str): switch between a matplotlib or a plotly render

    Returns:
    - None: The function displays the plot.

    """
    pridiction_q = get_quantiles(prediction, lower_q, upper_q, lower_cut, upper_cut)
    d_data = measure_ts.to_frame()
    d_data["pred_low"] = pridiction_q[0, :]
    d_data["pred_med"] = pridiction_q[1, :]
    d_data["pred_up"] = pridiction_q[2, :]

    if backend == "plotly":
        fig = make_subplots()
        fig.add_trace(
            go.Scatter(
                x=d_data.index,
                y=d_data.iloc[:, 0],
                mode="markers",
                name="Observed",
                marker=dict(color="green", size=10),
            )
        )
        fig.add_trace(
            go.Scatter(
                x=d_data.index,
                y=d_data["pred_med"],
                mode="lines",
                name="Predicted Median",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=d_data.index,
                y=d_data["pred_low"],
                mode="lines",
                fill=None,
                line=dict(color="orange"),
                name=f"Quantile {lower_q}",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=d_data.index,
                y=d_data["pred_up"],
                mode="lines",
                fill="tonexty",
                line=dict(color="orange"),
                name=f"Quantile {upper_q}",
            )
        )
        fig.update_layout(title=title, xaxis_title="Time", yaxis_title=y_label)
        return fig

    elif backend == "matplotlib":
        plt.figure(figsize=figsize)
        plt.scatter(
            d_data.index,
            d_data.iloc[:, 0],
            color="green",
            label="Observed",
            alpha=0.5,
        )

        plt.plot(
            d_data.index,
            d_data["pred_med"],
            color="orange",
            label="Predicted Median",
        )

        plt.fill_between(
            d_data.index,
            d_data["pred_low"],
            d_data["pred_up"],
            color="orange",
            alpha=0.1,
        )

        plt.ylabel(y_label)
        plt.title(title)
        if image_path is not None:
            plt.savefig(image_path, format="png", bbox_inches="tight")
        return plt.gcf()

    else:
        raise ValueError(
            f"{backend} is an invalid backend argument, choose one of"
            f"'plotly' or 'matplotlib"
        )


def changepoint_graph(
    x_variable: pd.Series,
    y_measure: pd.Series,
    prediction: np.ndarray,
    changepoint_periods: pd.Series = None,
    lower_q=0.025,
    upper_q=0.975,
    x_label: str = None,
    y_label: str = None,
    title: str = None,
    upper_cut=None,
    lower_cut=None,
    image_path: Path = None,
    backend: str = "plotly",
):
    """
    Visualise target data, measures and prediction as a function of an independent
    variable. Adapted to change point model. Prediction is displayed as a surface
    bounded by the lower an upper quartiles. A surface is drawn for each changepoint
    period.

    Parameters:
    - x_variable (pd.Series): The independent variable data.
    - y_measure (pd.Series): The dependent variable data.
    - prediction (np.ndarray): Array containing predictions.
    - changepoint_periods (pd.Series, optional): Series containing changepoint periods
        (default is None, only one period is considered).
    - lower_q (float): Lower quantile for the high-density interval (default is 0.025).
    - upper_q (float): Upper quantile for the high-density interval (default is 0.975).
    - x_label (str, optional): Label for the x-axis (default is None).
    - y_label (str, optional): Label for the y-axis (default is None).
    - title (str, optional): Title of the plot (default is None).
    - upper_cut (float, optional): Upper bound for the high-density interval
        (default is None).
    - lower_cut (float, optional): Lower bound for the high-density interval
        (default is None).
    - backend (str, optional): Backend for plotting, choose either 'plotly' or
        'matplotlib' (default is 'plotly').

    Returns:
    - None: The function displays the plot.

    Raises:
    - ValueError: If the specified backend is not 'plotly' or 'matplotlib'.
    """

    if changepoint_periods is None:
        changepoint_periods = np.zeros(x_variable.shape[0])

    prediction_q = get_quantiles(prediction, lower_q, upper_q, lower_cut, upper_cut)
    d_data = pd.concat([x_variable, y_measure], axis=1)
    d_data["pred_low"] = prediction_q[0, :]
    d_data["pred_med"] = prediction_q[1, :]
    d_data["pred_up"] = prediction_q[2, :]

    x_name = x_variable.name
    y_name = y_measure.name

    d_data.sort_values(x_name, inplace=True)

    color_list = ["blue", "red", "orange", "green"]
    mask_list = []
    for period in set(changepoint_periods):
        mask_list.append(changepoint_periods == period)

    if backend == "plotly":
        fig = make_subplots()
        fig.add_trace(
            go.Scatter(
                x=d_data[x_name],
                y=d_data[y_name],
                mode="markers",
                marker=dict(color=changepoint_periods, colorscale="Bluered", size=10),
                name="Observed",
            )
        )
        for mask, color in zip(mask_list, color_list):
            fig.add_trace(
                go.Scatter(
                    x=d_data.loc[mask, x_name],
                    y=d_data.loc[mask, "pred_med"],
                    mode="lines",
                    line=dict(color=color),
                    name="Predicted Median",
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=d_data.loc[mask, x_name],
                    y=d_data.loc[mask, "pred_low"],
                    mode="lines",
                    fill=None,
                    line=dict(color=color),
                    name="Lower Bound",
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=d_data.loc[mask, x_name],
                    y=d_data.loc[mask, "pred_up"],
                    mode="lines",
                    fill="tonexty",
                    line=dict(color=color),
                    name="Upper Bound",
                )
            )
        fig.update_layout(title=title, xaxis_title=x_label, yaxis_title=y_label)
        return fig

    elif backend == "matplotlib":
        plt.figure(figsize=(10, 6))
        plt.scatter(
            d_data[x_name],
            d_data[y_name],
            c=changepoint_periods,
            cmap="coolwarm",
            label="Observed",
            alpha=0.5,
        )

        for mask, color in zip(mask_list, color_list):
            plt.plot(
                d_data.loc[mask, x_name],
                d_data.loc[mask, "pred_med"],
                color=color,
                label="Predicted Median",
            )

            plt.fill_between(
                d_data.loc[mask, x_name],
                d_data.loc[mask, "pred_low"],
                d_data.loc[mask, "pred_up"],
                color=color,
                alpha=0.1,
            )

        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.title(title)
        if image_path is not None:
            plt.savefig(image_path, format="png", bbox_inches="tight")
        return plt.gcf()

    else:
        raise ValueError(
            f"{backend} is an invalid backend argument, choose one of"
            f"'plotly' or 'matplotlib"
        )


def time_series_bar_plot(
    measure: pd.Series,
    prediction: np.ndarray | xarray.core.dataarray.DataArray,
    lower_q=0.025,
    upper_q=0.975,
    upper_cut=None,
    lower_cut=None,
    bar_width=0.35,
    title: str = None,
    y_label: str = None,
    image_path: Path = None,
):
    """
    Generate a time series bar plot comparing a measure against a prediction with
    error bars representing quantiles.

    Args:
        measure (pd.Series): The measured values as a Pandas Series.
        prediction (np.ndarray or xarray.core.dataarray.DataArray):
            The predicted values or data array.
        lower_q (float, optional): The lower quantile for the error bars.
        Defaults to 0.025.
        upper_q (float, optional): The upper quantile for the error bars.
        Defaults to 0.975.
        upper_cut (float, optional): Upper limit for outliers. Defaults to None.
        lower_cut (float, optional): Lower limit for outliers. Defaults to None.
        bar_width (float, optional): Width of each bar. Defaults to 0.35.
        title (str, optional): Title for the plot. Defaults to None.
        y_label (str, optional): Label for the y-axis. Defaults to None.
    """

    lower_q, med, upper_q = get_quantiles(
        prediction, lower_q, upper_q, lower_cut, upper_cut
    )

    index = range(len(measure.index))

    plt.figure(figsize=(6, 5))
    plt.bar(
        [i - bar_width / 2 for i in index],
        med,
        width=bar_width,
        yerr=[med - lower_q, upper_q - med],
        capsize=10,
        label="Modèle",
    )
    plt.bar(
        [i + bar_width / 2 for i in index], measure, width=bar_width, label="Mesure"
    )
    plt.title(title)
    plt.ylabel(y_label)
    plt.xticks(index, measure.index)
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter("%m-%d"))
    plt.tight_layout()
    plt.legend()
    if image_path is not None:
        plt.savefig(image_path, format="png", bbox_inches="tight")
    return plt.gcf()


def compare_bars(
    measure: float,
    prediction: np.ndarray | xarray.core.dataarray.DataArray,
    lower_q: float = 0.025,
    upper_q: float = 0.975,
    upper_cut: int | float = None,
    lower_cut: int | float = None,
    title: str = None,
    y_label: str = None,
    measure_label: str = "Measure",
    prediction_label: str = "Model",
    image_path: Path = None,
    font_size: int = 12,
):
    """
    Compare a measure against a prediction using bar plots with error bars
    representing quantiles.

    Args:
        measure (float): The measured value to compare against the prediction.
        prediction (np.ndarray or xarray.core.dataarray.DataArray):
            The predicted values or data array.
        lower_q (float, optional): The lower quantile for the error bars.
        Defaults to 0.025.
        upper_q (float, optional): The upper quantile for the error bars.
        Defaults to 0.975.
        upper_cut (float, optional): Upper limit for outliers. Defaults to None.
        lower_cut (float, optional): Lower limit for outliers. Defaults to None.
        title (str, optional): Title for the plot. Defaults to None.
        y_label (str, optional): Label for the y-axis. Defaults to None.
        measure_label (str, optional): Label for the measure. Defaults to "Measure".
        prediction_label (str, optional): Label for the prediction. Defaults to "Model".
        image_path (Path, optional): Saving path to png image
        font_size (Int, optional): Fontsize for all figure text. Default 12
    """

    lower_q, med, upper_q = get_quantiles(
        prediction, lower_q, upper_q, lower_cut, upper_cut
    )

    plt.figure(figsize=(6, 5))
    plt.bar(
        0,
        med,
        yerr=[[med - lower_q], [upper_q - med]],
        capsize=10,
        label=prediction_label,
    )
    plt.bar(1, measure, label=measure_label)
    plt.title(title, fontsize=font_size)  # Set font size for title
    plt.ylabel(y_label, fontsize=font_size)  # Set font size for y-label
    plt.xticks(
        [0, 1], [prediction_label, measure_label], fontsize=font_size
    )  # Set font size for x-ticks
    plt.yticks(fontsize=font_size)  # Set font size for y-ticks

    if image_path is not None:
        plt.savefig(image_path, format="png", bbox_inches="tight")

    return plt.gcf()
